@@ -18,9 +18,9 @@ import time
 import logging
 import pprint
 import shutil
-from collections import OrderedDict
 
 from tools import building
+from tools import cache
 from tools import diagnostics
 from tools import js_manipulation
 from tools import shared
@@ -89,6 +89,7 @@ def optimize_syscalls(declares):
       'fd_seek',
       'fd_write',
       'fd_close',
+      'fd_fdstat_get',
     }):
       if DEBUG:
         logger.debug('very limited syscalls (%s) so disabling full filesystem support', ', '.join(str(s) for s in syscalls))
@@ -111,42 +112,62 @@ def to_nice_ident(ident): # limited version of the JS function toNiceIdent
   return ident.replace('%', '$').replace('@', '_').replace('.', '_')
 
 
-def update_settings_glue(metadata):
-  optimize_syscalls(metadata['declares'])
+def get_weak_imports(main_wasm):
+  dylink_sec = webassembly.parse_dylink_section(main_wasm)
+  for symbols in dylink_sec.import_info.values():
+    for symbol, flags in symbols.items():
+      if flags & webassembly.SYMBOL_BINDING_MASK == webassembly.SYMBOL_BINDING_WEAK:
+        settings.WEAK_IMPORTS.append(symbol)
+
+
+def update_settings_glue(wasm_file, metadata):
+  optimize_syscalls(metadata.imports)
 
   # Integrate info from backend
   if settings.SIDE_MODULE:
     # we don't need any JS library contents in side modules
     settings.DEFAULT_LIBRARY_FUNCS_TO_INCLUDE = []
   else:
-    syms = settings.DEFAULT_LIBRARY_FUNCS_TO_INCLUDE + [to_nice_ident(d) for d in metadata['declares']]
-    syms = set(syms).difference(metadata['exports'])
-    syms.update(metadata['globalImports'])
+    syms = settings.DEFAULT_LIBRARY_FUNCS_TO_INCLUDE + [to_nice_ident(d) for d in metadata.imports]
+    syms = set(syms).difference(metadata.exports)
     settings.DEFAULT_LIBRARY_FUNCS_TO_INCLUDE = sorted(syms)
+    if settings.MAIN_MODULE:
+      get_weak_imports(wasm_file)
 
-  settings.WASM_EXPORTS = metadata['exports'] + list(metadata['namedGlobals'].keys())
+  settings.WASM_EXPORTS = metadata.exports + list(metadata.namedGlobals.keys())
   # Store function exports so that Closure and metadce can track these even in
   # -sDECLARE_ASM_MODULE_EXPORTS=0 builds.
-  settings.WASM_FUNCTION_EXPORTS = metadata['exports']
+  settings.WASM_FUNCTION_EXPORTS = metadata.exports
 
   # start with the MVP features, and add any detected features.
-  settings.BINARYEN_FEATURES = ['--mvp-features'] + metadata['features']
+  settings.BINARYEN_FEATURES = ['--mvp-features'] + metadata.features
+  if settings.ASYNCIFY == 2:
+    settings.BINARYEN_FEATURES += ['--enable-reference-types']
+
   if settings.USE_PTHREADS:
     assert '--enable-threads' in settings.BINARYEN_FEATURES
   if settings.MEMORY64:
     assert '--enable-memory64' in settings.BINARYEN_FEATURES
 
-  settings.HAS_MAIN = bool(settings.MAIN_MODULE) or settings.STANDALONE_WASM or 'main' in settings.WASM_EXPORTS
+  settings.HAS_MAIN = bool(settings.MAIN_MODULE) or settings.PROXY_TO_PTHREAD or settings.STANDALONE_WASM or 'main' in settings.WASM_EXPORTS or '__main_argc_argv' in settings.WASM_EXPORTS
+  if settings.HAS_MAIN and not settings.MINIMAL_RUNTIME:
+    # Dependencies of `callMain`
+    settings.DEFAULT_LIBRARY_FUNCS_TO_INCLUDE += ['$exitJS', '$handleException']
 
   # When using dynamic linking the main function might be in a side module.
   # To be safe assume they do take input parametes.
-  settings.MAIN_READS_PARAMS = metadata['mainReadsParams'] or bool(settings.MAIN_MODULE)
+  settings.MAIN_READS_PARAMS = metadata.mainReadsParams or bool(settings.MAIN_MODULE)
+  if settings.MAIN_READS_PARAMS and not settings.STANDALONE_WASM:
+    # callMain depends on this library function
+    settings.DEFAULT_LIBRARY_FUNCS_TO_INCLUDE += ['$allocateUTF8OnStack']
 
   if settings.STACK_OVERFLOW_CHECK and not settings.SIDE_MODULE:
-    settings.EXPORTED_RUNTIME_METHODS += ['writeStackCookie', 'checkStackCookie']
     # writeStackCookie and checkStackCookie both rely on emscripten_stack_get_end being
     # exported.  In theory it should always be present since its defined in compiler-rt.
-    assert 'emscripten_stack_get_end' in metadata['exports']
+    assert 'emscripten_stack_get_end' in metadata.exports
+
+  for deps in metadata.jsDeps:
+    settings.DEFAULT_LIBRARY_FUNCS_TO_INCLUDE.extend(deps.split(','))
 
 
 def apply_static_code_hooks(forwarded_json, code):
@@ -180,7 +201,7 @@ def compile_settings():
     env['EMCC_BUILD_DIR'] = os.getcwd()
     out = shared.run_js_tool(path_from_root('src/compiler.js'),
                              [settings_file], stdout=subprocess.PIPE, stderr=stderr_file,
-                             cwd=path_from_root('src'), env=env)
+                             cwd=path_from_root('src'), env=env, encoding='utf-8')
   assert '//FORWARDED_DATA:' in out, 'Did not receive forwarded data in pre output - process failed?'
   glue, forwarded_data = out.split('//FORWARDED_DATA:')
   return glue, forwarded_data
@@ -188,9 +209,9 @@ def compile_settings():
 
 def set_memory(static_bump):
   stack_low = align_memory(settings.GLOBAL_BASE + static_bump)
-  stack_high = align_memory(stack_low + settings.TOTAL_STACK)
-  settings.STACK_BASE = stack_high
-  settings.STACK_MAX = stack_low
+  stack_high = align_memory(stack_low + settings.STACK_SIZE)
+  settings.STACK_HIGH = stack_high
+  settings.STACK_LOW = stack_low
   settings.HEAP_BASE = align_memory(stack_high)
 
 
@@ -209,19 +230,12 @@ def report_missing_symbols(js_library_funcs):
     # In this mode we never expect _main in the export list.
     return
 
-  # PROXY_TO_PTHREAD only makes sense with a main(), so error if one is
-  # missing. note that when main() might arrive from another module we cannot
-  # error here.
-  if settings.PROXY_TO_PTHREAD and '_main' not in defined_symbols and \
-     not settings.RELOCATABLE:
-    exit_with_error('PROXY_TO_PTHREAD proxies main() for you, but no main exists')
-
   if settings.IGNORE_MISSING_MAIN:
     # The default mode for emscripten is to ignore the missing main function allowing
     # maximum compatibility.
     return
 
-  if settings.EXPECT_MAIN and 'main' not in settings.WASM_EXPORTS:
+  if settings.EXPECT_MAIN and 'main' not in settings.WASM_EXPORTS and '__main_argc_argv' not in settings.WASM_EXPORTS:
     # For compatibility with the output of wasm-ld we use the same wording here in our
     # error message as if wasm-ld had failed (i.e. in LLD_REPORT_UNDEFINED mode).
     exit_with_error('entry symbol not defined (pass --no-entry to suppress): main')
@@ -268,7 +282,7 @@ def trim_asm_const_body(body):
 
 def create_named_globals(metadata):
   named_globals = []
-  for k, v in metadata['namedGlobals'].items():
+  for k, v in metadata.namedGlobals.items():
     v = int(v)
     if settings.RELOCATABLE:
       v += settings.GLOBAL_BASE
@@ -296,12 +310,36 @@ def emscript(in_wasm, out_wasm, outfile_js, memfile):
 
   metadata = finalize_wasm(in_wasm, out_wasm, memfile)
 
-  update_settings_glue(metadata)
+  if settings.RELOCATABLE and settings.MEMORY64 == 2:
+    metadata.imports += ['__memory_base32']
+
+  update_settings_glue(out_wasm, metadata)
+
+  if not settings.WASM_BIGINT and metadata.emJsFuncs:
+    import_map = {}
+
+    with webassembly.Module(in_wasm) as module:
+      types = module.get_types()
+      for imp in module.get_imports():
+        import_map[imp.field] = imp
+
+    for em_js_func, raw in metadata.emJsFuncs.items():
+      c_sig = raw.split('<::>')[0].strip('()')
+      if not c_sig or c_sig == 'void':
+        c_sig = []
+      else:
+        c_sig = c_sig.split(',')
+      if em_js_func in import_map:
+        imp = import_map[em_js_func]
+        assert imp.kind == webassembly.ExternType.FUNC
+        signature = types[imp.type]
+        if len(signature.params) != len(c_sig):
+          diagnostics.warning('em-js-i64', 'using 64-bit arguments in EM_JS function without WASM_BIGINT is not yet fully supported: `%s` (%s, %s)', em_js_func, c_sig, signature.params)
 
   if settings.SIDE_MODULE:
-    if metadata['asmConsts']:
+    if metadata.asmConsts:
       exit_with_error('EM_ASM is not supported in side modules')
-    if metadata['emJsFuncs']:
+    if metadata.emJsFuncs:
       exit_with_error('EM_JS is not supported in side modules')
     logger.debug('emscript: skipping remaining js glue generation')
     return
@@ -316,7 +354,7 @@ def emscript(in_wasm, out_wasm, outfile_js, memfile):
     dylink_sec = webassembly.parse_dylink_section(in_wasm)
     static_bump = align_memory(dylink_sec.mem_size)
     set_memory(static_bump)
-    logger.debug('stack_base: %d, stack_max: %d, heap_base: %d', settings.STACK_BASE, settings.STACK_MAX, settings.HEAP_BASE)
+    logger.debug('stack_low: %d, stack_high: %d, heap_base: %d', settings.STACK_LOW, settings.STACK_HIGH, settings.HEAP_BASE)
 
     # When building relocatable output (e.g. MAIN_MODULE) the reported table
     # size does not include the reserved slot at zero for the null pointer.
@@ -324,6 +362,13 @@ def emscript(in_wasm, out_wasm, outfile_js, memfile):
     if settings.INITIAL_TABLE == -1:
       settings.INITIAL_TABLE = dylink_sec.table_size + 1
 
+    if settings.ASYNCIFY:
+      metadata.imports += ['__asyncify_state', '__asyncify_data']
+
+  invoke_funcs = metadata.invokeFuncs
+  if invoke_funcs:
+    settings.DEFAULT_LIBRARY_FUNCS_TO_INCLUDE += ['$getWasmTableEntry']
+
   glue, forwarded_data = compile_settings()
   if DEBUG:
     logger.debug('  emscript: glue took %s seconds' % (time.time() - t))
@@ -331,6 +376,9 @@ def emscript(in_wasm, out_wasm, outfile_js, memfile):
 
   forwarded_json = json.loads(forwarded_data)
 
+  if forwarded_json['warnings']:
+    diagnostics.warning('js-compiler', 'warnings in JS library compilation')
+
   pre, post = glue.split('// EMSCRIPTEN_END_FUNCS')
 
   if settings.ASSERTIONS:
@@ -340,12 +388,12 @@ def emscript(in_wasm, out_wasm, outfile_js, memfile):
         pre += f"  ignoredModuleProp('{sym}');\n"
     pre += "}\n"
 
-  exports = metadata['exports']
+  exports = metadata.exports
 
   if settings.ASYNCIFY:
     exports += ['asyncify_start_unwind', 'asyncify_stop_unwind', 'asyncify_start_rewind', 'asyncify_stop_rewind']
 
-  report_missing_symbols(forwarded_json['libraryFunctions'])
+  report_missing_symbols(forwarded_json['librarySymbols'])
 
   if not outfile_js:
     logger.debug('emscript: skipping remaining js glue generation')
@@ -367,11 +415,10 @@ def emscript(in_wasm, out_wasm, outfile_js, memfile):
     ('// === Body ===\n\n' + asm_const_map +
      '\n'.join(em_js_funcs) + '\n'))
 
-  with open(outfile_js, 'w') as out:
+  with open(outfile_js, 'w', encoding='utf-8') as out:
     out.write(normalize_line_endings(pre))
     pre = None
 
-    invoke_funcs = metadata['invokeFuncs']
     sending = create_sending(invoke_funcs, metadata)
     receiving = create_receiving(exports)
 
@@ -390,25 +437,12 @@ def emscript(in_wasm, out_wasm, outfile_js, memfile):
 
 def remove_trailing_zeros(memfile):
   mem_data = utils.read_binary(memfile)
-  end = len(mem_data)
-  while end > 0 and (mem_data[end - 1] == b'\0' or mem_data[end - 1] == 0):
-    end -= 1
-  utils.write_binary(memfile, mem_data[:end])
+  mem_data = mem_data.rstrip(b'\0')
+  utils.write_binary(memfile, mem_data)
 
 
 @ToolchainProfiler.profile()
-def get_metadata_binaryen(infile, outfile, modify_wasm, args):
-  stdout = building.run_binaryen_command('wasm-emscripten-finalize',
-                                         infile=infile,
-                                         outfile=outfile if modify_wasm else None,
-                                         args=args,
-                                         stdout=subprocess.PIPE)
-  metadata = load_metadata_json(stdout)
-  return metadata
-
-
-@ToolchainProfiler.profile()
-def get_metadata_python(infile, outfile, modify_wasm, args):
+def get_metadata(infile, outfile, modify_wasm, args):
   metadata = extract_metadata.extract_metadata(infile)
   if modify_wasm:
     # In some cases we still need to modify the wasm file
@@ -416,44 +450,15 @@ def get_metadata_python(infile, outfile, modify_wasm, args):
     building.run_binaryen_command('wasm-emscripten-finalize',
                                   infile=infile,
                                   outfile=outfile,
-                                  args=args,
-                                  stdout=subprocess.PIPE)
+                                  args=args)
     # When we do this we can generate new imports, so
     # re-read parts of the metadata post-finalize
     extract_metadata.update_metadata(outfile, metadata)
-  elif 'main' in metadata['exports']:
-    # Mimic a bug in wasm-emscripten-finalize where we don't correctly
-    # detect the presense of the main wrapper function unless we are
-    # modifying the binary.  This is because binaryen doesn't reaad
-    # the function bodies in this mode.
-    # TODO(sbc): Remove this once we make the switch away from
-    # binaryen metadata.
-    metadata['mainReadsParams'] = 1
+  if DEBUG:
+    logger.debug("Metadata: " + pprint.pformat(metadata.__dict__))
   return metadata
 
 
-# Test function for comparing binaryen vs python metadata.
-# Remove this once we go back to having just one method.
-def compare_metadata(metadata, pymetadata):
-  if sorted(metadata.keys()) != sorted(pymetadata.keys()):
-    print(sorted(metadata.keys()))
-    print(sorted(pymetadata.keys()))
-    exit_with_error('metadata keys mismatch')
-  for key in metadata:
-    old = metadata[key]
-    new = pymetadata[key]
-    if key == 'features':
-      old = sorted(old)
-      new = sorted(new)
-    if old != new:
-      print(key)
-      open(path_from_root('first.txt'), 'w').write(pprint.pformat(old))
-      open(path_from_root('second.txt'), 'w').write(pprint.pformat(new))
-      print(pprint.pformat(old))
-      print(pprint.pformat(new))
-      exit_with_error('metadata mismatch')
-
-
 def finalize_wasm(infile, outfile, memfile):
   building.save_intermediate(infile, 'base.wasm')
   args = []
@@ -491,15 +496,19 @@ def finalize_wasm(infile, outfile, memfile):
       args.append('--dyncalls-i64')
       # we need to add some dyncalls to the wasm
       modify_wasm = True
-  if settings.LEGALIZE_JS_FFI:
-    # When we dynamically link our JS loader adds functions from wasm modules to
-    # the table. It must add the original versions of them, not legalized ones,
-    # so that indirect calls have the right type, so export those.
-    if settings.RELOCATABLE:
-      args.append('--pass-arg=legalize-js-interface-export-originals')
-    modify_wasm = True
+  if settings.AUTODEBUG:
+    # In AUTODEBUG mode we want to delay all legalization until later.  This is hack
+    # to force wasm-emscripten-finalize not to do any legalization at all.
+    args.append('--bigint')
   else:
-    args.append('--no-legalize-javascript-ffi')
+    if settings.LEGALIZE_JS_FFI:
+      # When we dynamically link our JS loader adds functions from wasm modules to
+      # the table. It must add the original versions of them, not legalized ones,
+      # so that indirect calls have the right type, so export those.
+      args += building.js_legalization_pass_flags()
+      modify_wasm = True
+    else:
+      args.append('--no-legalize-javascript-ffi')
   if memfile:
     args.append(f'--separate-data-segments={memfile}')
     args.append(f'--global-base={settings.GLOBAL_BASE}')
@@ -515,30 +524,7 @@ def finalize_wasm(infile, outfile, memfile):
   if settings.DEBUG_LEVEL >= 3:
     args.append('--dwarf')
 
-  # Currently we have two different ways to extract the metadata from the
-  # wasm binary:
-  # 1. via wasm-emscripten-finalize (binaryen)
-  # 2. via local python code
-  # We also have a 'compare' mode that runs both extraction methods and
-  # checks that they produce identical results.
-  read_metadata = os.environ.get('EMCC_READ_METADATA', 'binaryen')
-  if read_metadata == 'binaryen':
-    metadata = get_metadata_binaryen(infile, outfile, modify_wasm, args)
-  elif read_metadata == 'python':
-    metadata = get_metadata_python(infile, outfile, modify_wasm, args)
-  elif read_metadata == 'compare':
-    shutil.copy2(infile, infile + '.bak')
-    if settings.GENERATE_SOURCE_MAP:
-      shutil.copy2(infile + '.map', infile + '.map.bak')
-    pymetadata = get_metadata_python(infile, outfile, modify_wasm, args)
-    shutil.move(infile + '.bak', infile)
-    if settings.GENERATE_SOURCE_MAP:
-      shutil.move(infile + '.map.bak', infile + '.map')
-    metadata = get_metadata_binaryen(infile, outfile, modify_wasm, args)
-    compare_metadata(metadata, pymetadata)
-  else:
-    assert False
-
+  metadata = get_metadata(infile, outfile, modify_wasm, args)
   if modify_wasm:
     building.save_intermediate(infile, 'post_finalize.wasm')
   elif infile != outfile:
@@ -553,13 +539,25 @@ def finalize_wasm(infile, outfile, memfile):
     # the dynamic linking case, our loader zeros it out)
     remove_trailing_zeros(memfile)
 
+  expected_exports = set(settings.EXPORTED_FUNCTIONS)
+  expected_exports.update(asmjs_mangle(s) for s in settings.REQUIRED_EXPORTS)
+
+  # Calculate the subset of exports that were explicitly marked with llvm.used.
+  # These are any exports that were not requested on the command line and are
+  # not known auto-generated system functions.
+  unexpected_exports = [e for e in metadata.exports if treat_as_user_function(e)]
+  unexpected_exports = [asmjs_mangle(e) for e in unexpected_exports]
+  unexpected_exports = [e for e in unexpected_exports if e not in expected_exports]
+  building.user_requested_exports.update(unexpected_exports)
+  settings.EXPORTED_FUNCTIONS.extend(unexpected_exports)
+
   return metadata
 
 
 def create_asm_consts(metadata):
   asm_consts = {}
-  for addr, const in metadata['asmConsts'].items():
-    const = trim_asm_const_body(const)
+  for addr, const in metadata.asmConsts.items():
+    body = trim_asm_const_body(const)
     args = []
     max_arity = 16
     arity = 0
@@ -568,17 +566,46 @@ def create_asm_consts(metadata):
         arity = i + 1
     for i in range(arity):
       args.append('$' + str(i))
-    const = 'function(' + ', '.join(args) + ') {' + const + '}'
-    asm_consts[int(addr)] = const
+    args = ', '.join(args)
+    if 'arguments' in body:
+      # arrow functions don't bind `arguments` so we have to use
+      # the old function syntax in this case
+      func = f'function({args}) {{ {body} }}'
+    else:
+      func = f'({args}) => {{ {body} }}'
+    if settings.RELOCATABLE:
+      addr += settings.GLOBAL_BASE
+    asm_consts[addr] = func
   asm_consts = [(key, value) for key, value in asm_consts.items()]
   asm_consts.sort()
   return asm_consts
 
 
+def type_to_sig(type):
+  # These should match the conversion in $sigToWasmTypes.
+  return {
+    webassembly.Type.I32: 'i',
+    webassembly.Type.I64: 'j',
+    webassembly.Type.F32: 'f',
+    webassembly.Type.F64: 'd',
+    webassembly.Type.VOID: 'v'
+  }[type]
+
+
+def func_type_to_sig(type):
+  parameters = [type_to_sig(param) for param in type.params]
+  if type.returns:
+    assert len(type.returns) == 1, "One return type expected."
+    ret = type.returns[0]
+  else:
+    ret = webassembly.Type.VOID
+  return type_to_sig(ret) + ''.join(parameters)
+
+
 def create_em_js(metadata):
   em_js_funcs = []
   separator = '<::>'
-  for name, raw in metadata.get('emJsFuncs', {}).items():
+  for name, raw in metadata.emJsFuncs.items():
     assert separator in raw
     args, body = raw.split(separator, 1)
     args = args[1:-1]
@@ -587,13 +614,19 @@ def create_em_js(metadata):
     else:
       args = args.split(',')
     arg_names = [arg.split()[-1].replace("*", "") for arg in args if arg]
-    func = 'function {}({}){}'.format(name, ','.join(arg_names), body)
+    args = ','.join(arg_names)
+    func = f'function {name}({args}) {body}'
+    if settings.ASYNCIFY == 2 and name in metadata.emJsFuncTypes:
+      sig = func_type_to_sig(metadata.emJsFuncTypes[name])
+      func = func + f'\n{name}.sig = \'{sig}\';'
     em_js_funcs.append(func)
 
   return em_js_funcs
 
 
 def add_standard_wasm_imports(send_items_map):
+  extra_sent_items = []
+
   if settings.IMPORTED_MEMORY:
     memory_import = 'wasmMemory'
     if settings.MODULARIZE and settings.USE_PTHREADS:
@@ -603,143 +636,53 @@ def add_standard_wasm_imports(send_items_map):
     send_items_map['memory'] = memory_import
 
   if settings.SAFE_HEAP:
-    send_items_map['segfault'] = 'segfault'
-    send_items_map['alignfault'] = 'alignfault'
+    extra_sent_items.append('segfault')
+    extra_sent_items.append('alignfault')
 
   if settings.RELOCATABLE:
     send_items_map['__indirect_function_table'] = 'wasmTable'
-    if settings.EXCEPTION_HANDLING:
-      send_items_map['__cpp_exception'] = '___cpp_exception'
-    if settings.SUPPORT_LONGJMP == 'wasm':
-      send_items_map['__c_longjmp'] = '___c_longjmp'
-
-  if settings.MAYBE_WASM2JS or settings.AUTODEBUG or settings.LINKABLE:
-    # legalization of i64 support code may require these in some modes
-    send_items_map['setTempRet0'] = 'setTempRet0'
-    send_items_map['getTempRet0'] = 'getTempRet0'
 
   if settings.AUTODEBUG:
-    send_items_map['log_execution'] = '''function(loc) {
-      console.log('log_execution ' + loc);
-    }'''
-    send_items_map['get_i32'] = '''function(loc, index, value) {
-      console.log('get_i32 ' + [loc, index, value]);
-      return value;
-    }'''
-    send_items_map['get_i64'] = '''function(loc, index, low, high) {
-      console.log('get_i64 ' + [loc, index, low, high]);
-      setTempRet0(high);
-      return low;
-    }'''
-    send_items_map['get_f32'] = '''function(loc, index, value) {
-      console.log('get_f32 ' + [loc, index, value]);
-      return value;
-    }'''
-    send_items_map['get_f64'] = '''function(loc, index, value) {
-      console.log('get_f64 ' + [loc, index, value]);
-      return value;
-    }'''
-    send_items_map['get_anyref'] = '''function(loc, index, value) {
-      console.log('get_anyref ' + [loc, index, value]);
-      return value;
-    }'''
-    send_items_map['get_exnref'] = '''function(loc, index, value) {
-      console.log('get_exnref ' + [loc, index, value]);
-      return value;
-    }'''
-    send_items_map['set_i32'] = '''function(loc, index, value) {
-      console.log('set_i32 ' + [loc, index, value]);
-      return value;
-    }'''
-    send_items_map['set_i64'] = '''function(loc, index, low, high) {
-      console.log('set_i64 ' + [loc, index, low, high]);
-      setTempRet0(high);
-      return low;
-    }'''
-    send_items_map['set_f32'] = '''function(loc, index, value) {
-      console.log('set_f32 ' + [loc, index, value]);
-      return value;
-    }'''
-    send_items_map['set_f64'] = '''function(loc, index, value) {
-      console.log('set_f64 ' + [loc, index, value]);
-      return value;
-    }'''
-    send_items_map['set_anyref'] = '''function(loc, index, value) {
-      console.log('set_anyref ' + [loc, index, value]);
-      return value;
-    }'''
-    send_items_map['set_exnref'] = '''function(loc, index, value) {
-      console.log('set_exnref ' + [loc, index, value]);
-      return value;
-    }'''
-    send_items_map['load_ptr'] = '''function(loc, bytes, offset, ptr) {
-      console.log('load_ptr ' + [loc, bytes, offset, ptr]);
-      return ptr;
-    }'''
-    send_items_map['load_val_i32'] = '''function(loc, value) {
-      console.log('load_val_i32 ' + [loc, value]);
-      return value;
-    }'''
-    send_items_map['load_val_i64'] = '''function(loc, low, high) {
-      console.log('load_val_i64 ' + [loc, low, high]);
-      setTempRet0(high);
-      return low;
-    }'''
-    send_items_map['load_val_f32'] = '''function(loc, value) {
-      console.log('load_val_f32 ' + [loc, value]);
-      return value;
-    }'''
-    send_items_map['load_val_f64'] = '''function(loc, value) {
-      console.log('load_val_f64 ' + [loc, value]);
-      return value;
-    }'''
-    send_items_map['store_ptr'] = '''function(loc, bytes, offset, ptr) {
-      console.log('store_ptr ' + [loc, bytes, offset, ptr]);
-      return ptr;
-    }'''
-    send_items_map['store_val_i32'] = '''function(loc, value) {
-      console.log('store_val_i32 ' + [loc, value]);
-      return value;
-    }'''
-    send_items_map['store_val_i64'] = '''function(loc, low, high) {
-      console.log('store_val_i64 ' + [loc, low, high]);
-      setTempRet0(high);
-      return low;
-    }'''
-    send_items_map['store_val_f32'] = '''function(loc, value) {
-      console.log('store_val_f32 ' + [loc, value]);
-      return value;
-    }'''
-    send_items_map['store_val_f64'] = '''function(loc, value) {
-      console.log('store_val_f64 ' + [loc, value]);
-      return value;
-    }'''
+    extra_sent_items += [
+      'log_execution',
+      'get_i32',
+      'get_i64',
+      'get_f32',
+      'get_f64',
+      'get_anyref',
+      'get_exnref',
+      'set_i32',
+      'set_i64',
+      'set_f32',
+      'set_f64',
+      'set_anyref',
+      'set_exnref',
+      'load_ptr',
+      'load_val_i32',
+      'load_val_i64',
+      'load_val_f32',
+      'load_val_f64',
+      'store_ptr',
+      'store_val_i32',
+      'store_val_i64',
+      'store_val_f32',
+      'store_val_f64',
+    ]
+
+  for s in extra_sent_items:
+    send_items_map[s] = s
 
 
 def create_sending(invoke_funcs, metadata):
-  em_js_funcs = set(metadata['emJsFuncs'].keys())
-  declares = [asmjs_mangle(d) for d in metadata['declares']]
-  externs = [asmjs_mangle(e) for e in metadata['globalImports']]
-  send_items = set(invoke_funcs + declares + externs)
-  send_items.update(em_js_funcs)
+  # Map of wasm imports to mangled/external/JS names
+  send_items_map = {}
 
-  def fix_import_name(g):
-    # Unlike fastcomp the wasm backend doesn't use the '_' prefix for native
-    # symbols.  Emscripten currently expects symbols to start with '_' so we
-    # artificially add them to the output of emscripten-wasm-finalize and them
-    # strip them again here.
-    # note that we don't do this for EM_JS functions (which, rarely, may have
-    # a '_' prefix)
-    if g.startswith('_') and g not in em_js_funcs:
-      return g[1:]
-    return g
-
-  send_items_map = OrderedDict()
-  for name in send_items:
-    internal_name = fix_import_name(name)
-    if internal_name in send_items_map:
-      exit_with_error('duplicate symbol in exports to wasm: %s', name)
-    send_items_map[internal_name] = name
+  for name in metadata.emJsFuncs:
+    send_items_map[name] = name
+  for name in invoke_funcs:
+    send_items_map[name] = name
+  for name in metadata.imports:
+    send_items_map[name] = asmjs_mangle(name)
 
   add_standard_wasm_imports(send_items_map)
 
@@ -750,11 +693,16 @@ def create_sending(invoke_funcs, metadata):
 def make_export_wrappers(exports, delay_assignment):
   wrappers = []
   for name in exports:
+    # Tags cannot be wrapped in createExportWrapper
+    if name == '__cpp_exception':
+      continue
     mangled = asmjs_mangle(name)
     # The emscripten stack functions are called very early (by writeStackCookie) before
     # the runtime is initialized so we can't create these wrappers that check for
     # runtimeInitialized.
-    if settings.ASSERTIONS and not name.startswith('emscripten_stack_'):
+    # Likewise `__trap` can occur before the runtime is initialized since it is used in
+    # abort.
+    if settings.ASSERTIONS and not name.startswith('emscripten_stack_') and name != '__trap':
       # With assertions enabled we create a wrapper that are calls get routed through, for
       # the lifetime of the program.
       if delay_assignment:
@@ -790,8 +738,6 @@ def create_receiving(exports):
   if not settings.DECLARE_ASM_MODULE_EXPORTS:
     return ''
 
-  exports_that_are_not_initializers = [x for x in exports if x != building.WASM_CALL_CTORS]
-
   receiving = []
 
   # with WASM_ASYNC_COMPILATION that asm object may not exist at this point in time
@@ -799,32 +745,21 @@ def create_receiving(exports):
   delay_assignment = settings.WASM_ASYNC_COMPILATION and not settings.MINIMAL_RUNTIME
   if not delay_assignment:
     if settings.MINIMAL_RUNTIME:
-      # In Wasm exports are assigned inside a function to variables existing in top level JS scope, i.e.
+      # In Wasm exports are assigned inside a function to variables
+      # existing in top level JS scope, i.e.
       # var _main;
       # WebAssembly.instantiate(Module["wasm"], imports).then((function(output) {
       # var asm = output.instance.exports;
       # _main = asm["_main"];
       generate_dyncall_assignment = settings.DYNCALLS and '$dynCall' in settings.DEFAULT_LIBRARY_FUNCS_TO_INCLUDE
+      exports_that_are_not_initializers = [x for x in exports if x != building.WASM_CALL_CTORS]
+
       for s in exports_that_are_not_initializers:
         mangled = asmjs_mangle(s)
         dynCallAssignment = ('dynCalls["' + s.replace('dynCall_', '') + '"] = ') if generate_dyncall_assignment and mangled.startswith('dynCall_') else ''
         receiving += [dynCallAssignment + mangled + ' = asm["' + s + '"];']
     else:
-      if settings.MINIMAL_RUNTIME:
-        # In wasm2js exports can be directly processed at top level, i.e.
-        # var asm = Module["asm"](asmLibraryArg, buffer);
-        # var _main = asm["_main"];
-        if settings.USE_PTHREADS and settings.MODULARIZE:
-          # TODO: As a temp solution, multithreaded MODULARIZED MINIMAL_RUNTIME builds export all
-          # symbols like regular runtime does.
-          # Fix this by migrating worker.js code to reside inside the Module so it is in the same
-          # scope as the rest of the JS code, or by defining an export syntax to MINIMAL_RUNTIME
-          # that multithreaded MODULARIZEd builds can export on.
-          receiving += [asmjs_mangle(s) + ' = Module["' + asmjs_mangle(s) + '"] = asm["' + s + '"];' for s in exports_that_are_not_initializers]
-        else:
-          receiving += ['var ' + asmjs_mangle(s) + ' = asm["' + asmjs_mangle(s) + '"];' for s in exports_that_are_not_initializers]
-      else:
-        receiving += make_export_wrappers(exports, delay_assignment)
+      receiving += make_export_wrappers(exports, delay_assignment)
   else:
     receiving += make_export_wrappers(exports, delay_assignment)
 
@@ -840,7 +775,10 @@ def create_module(sending, receiving, invoke_funcs, metadata):
   module = []
 
   module.append('var asmLibraryArg = %s;\n' % sending)
-  if settings.ASYNCIFY and settings.ASSERTIONS:
+  if settings.ASYNCIFY and (settings.ASSERTIONS or settings.ASYNCIFY == 2):
+    # instrumenting imports is used in asyncify in two ways: to add assertions
+    # that check for proper import use, and for ASYNCIFY=2 we use them to set up
+    # the Promise API on the import side.
     module.append('Asyncify.instrumentWasmImports(asmLibraryArg);\n')
 
   if not settings.MINIMAL_RUNTIME:
@@ -848,51 +786,11 @@ def create_module(sending, receiving, invoke_funcs, metadata):
 
   module.append(receiving)
   module.append(invoke_wrappers)
+  if settings.MEMORY64:
+    module.append(create_wasm64_wrappers(metadata))
   return module
 
 
-def load_metadata_json(metadata_raw):
-  try:
-    metadata_json = json.loads(metadata_raw)
-  except Exception:
-    logger.error('emscript: failure to parse metadata output from wasm-emscripten-finalize. raw output is: \n' + metadata_raw)
-    raise
-
-  metadata = {
-    'declares': [],
-    'globalImports': [],
-    'exports': [],
-    'namedGlobals': {},
-    'emJsFuncs': {},
-    'asmConsts': {},
-    'invokeFuncs': [],
-    'features': [],
-    'mainReadsParams': 1,
-  }
-
-  for key, value in metadata_json.items():
-    if key not in metadata:
-      exit_with_error('unexpected metadata key received from wasm-emscripten-finalize: %s', key)
-    metadata[key] = value
-
-  if DEBUG:
-    logger.debug("Metadata parsed: " + pprint.pformat(metadata))
-
-  expected_exports = set(settings.EXPORTED_FUNCTIONS)
-  expected_exports.update(asmjs_mangle(s) for s in settings.REQUIRED_EXPORTS)
-
-  # Calculate the subset of exports that were explicitly marked with llvm.used.
-  # These are any exports that were not requested on the command line and are
-  # not known auto-generated system functions.
-  unexpected_exports = [e for e in metadata['exports'] if treat_as_user_function(e)]
-  unexpected_exports = [asmjs_mangle(e) for e in unexpected_exports]
-  unexpected_exports = [e for e in unexpected_exports if e not in expected_exports]
-  building.user_requested_exports.update(unexpected_exports)
-  settings.EXPORTED_FUNCTIONS.extend(unexpected_exports)
-
-  return metadata
-
-
 def create_invoke_wrappers(invoke_funcs):
   """Asm.js-style exception handling: invoke wrapper generation."""
   invoke_wrappers = ''
@@ -902,6 +800,66 @@ def create_invoke_wrappers(invoke_funcs):
   return invoke_wrappers
 
 
+def create_wasm64_wrappers(metadata):
+  # TODO(sbc): Move this into somewhere less static.  Maybe it can become
+  # part of library.js file, even though this metadata relates specifically
+  # to native (non-JS) functions.
+  #
+  # The signature format here is similar to the one used for JS libraries
+  # but with the following as the only valid char:
+  #  '_' - non-pointer argument (pass through unchanged)
+  #  'p' - pointer/int53 argument (convert to/from BigInt)
+  #  'P' - same as above but allow `undefined` too (requires extra check)
+  mapping = {
+    'sbrk': 'pP',
+    'stackAlloc': 'pp',
+    'emscripten_builtin_malloc': 'pp',
+    'malloc': 'pp',
+    '__getTypeName': 'pp',
+    'setThrew': '_p',
+    'free': '_p',
+    'stackRestore': '_p',
+    '__cxa_is_pointer_type': '_p',
+    'stackSave': 'p',
+    'fflush': '_p',
+    'emscripten_stack_get_end': 'p',
+    'emscripten_stack_get_base': 'p',
+    'pthread_self': 'p',
+    'emscripten_stack_get_current': 'p',
+    '__errno_location': 'p',
+    'emscripten_builtin_memalign': 'ppp',
+    'emscripten_builtin_free': 'vp',
+    'main': '__PP',
+    '__main_argc_argv': '__PP',
+    'emscripten_stack_set_limits': '_pp',
+    '__set_stack_limits': '_pp',
+    '__cxa_can_catch': '_ppp',
+    '_wasmfs_write_file': '_ppp',
+    '__dl_seterr': '_pp',
+  }
+
+  wasm64_wrappers = '''
+function instrumentWasmExportsForMemory64(exports) {
+  // First, make a copy of the incoming exports object
+  exports = Object.assign({}, exports);'''
+
+  sigs_seen = set()
+  wrap_functions = []
+  for exp in metadata.exports:
+    sig = mapping.get(exp)
+    if sig:
+      if sig not in sigs_seen:
+        sigs_seen.add(sig)
+        wasm64_wrappers += js_manipulation.make_wasm64_wrapper(sig)
+      wrap_functions.append(exp)
+
+  for f in wrap_functions:
+    sig = mapping[f]
+    wasm64_wrappers += f"\n  exports['{f}'] = wasm64Wrapper_{sig}(exports['{f}']);"
+  wasm64_wrappers += '\n  return exports\n}'
+  return wasm64_wrappers
+
+
 def normalize_line_endings(text):
   """Normalize to UNIX line endings.
 
@@ -913,8 +871,8 @@ def normalize_line_endings(text):
 
 
 def clear_struct_info():
-  output_name = shared.Cache.get_lib_name('struct_info.json', varies=False)
-  shared.Cache.erase_file(output_name)
+  output_name = cache.get_lib_name('struct_info.json', varies=False)
+  cache.erase_file(output_name)
 
 
 def generate_struct_info():
@@ -927,8 +885,8 @@ def generate_struct_info():
   def generate_struct_info(out):
     gen_struct_info.main(['-q', '-o', out])
 
-  output_name = shared.Cache.get_lib_name('struct_info.json', varies=False)
-  settings.STRUCT_INFO = shared.Cache.get(output_name, generate_struct_info)
+  output_name = cache.get_lib_name('struct_info.json', varies=False)
+  settings.STRUCT_INFO = cache.get(output_name, generate_struct_info)
 
 
 def run(in_wasm, out_wasm, outfile_js, memfile):
