@@ -17,6 +17,9 @@ global.proxiedFunctionTable = ['null'/* Reserve index 0 for an undefined functio
 
 // Mangles the given C/JS side function name to assembly level function name (adds an underscore)
 function mangleCSymbolName(f) {
+  if (f === '__main_argc_argv') {
+    f = 'main';
+  }
   return f[0] == '$' ? f.substr(1) : '_' + f;
 }
 
@@ -44,15 +47,17 @@ function stringifyWithFunctions(obj) {
   if (obj === null || typeof obj != 'object') return JSON.stringify(obj);
   if (Array.isArray(obj)) {
     return '[' + obj.map(stringifyWithFunctions).join(',') + ']';
-  } else {
-    return '{' + Object.keys(obj).map((key) => escapeJSONKey(key) + ':' + stringifyWithFunctions(obj[key])).join(',') + '}';
   }
+  return '{' + Object.keys(obj).map((key) => escapeJSONKey(key) + ':' + stringifyWithFunctions(obj[key])).join(',') + '}';
 }
 
 function isDefined(symName) {
   if (WASM_EXPORTS.has(symName) || SIDE_MODULE_EXPORTS.has(symName)) {
     return true;
   }
+  if (symName == '__main_argc_argv' && SIDE_MODULE_EXPORTS.has('main')) {
+    return true;
+  }
   // 'invoke_' symbols are created at runtime in libary_dylink.py so can
   // always be considered as defined.
   if (RELOCATABLE && symName.startsWith('invoke_')) {
@@ -75,6 +80,11 @@ function runJSify(functionsOnly) {
     LibraryManager.load();
 
     const libFuncsToInclude = DEFAULT_LIBRARY_FUNCS_TO_INCLUDE;
+    for (const sym of EXPORTED_RUNTIME_METHODS) {
+      if ('$' + sym in LibraryManager.library) {
+        libFuncsToInclude.push('$' + sym);
+      }
+    }
     if (INCLUDE_FULL_LIBRARY) {
       for (const key in LibraryManager.library) {
         if (!isJsLibraryConfigIdentifier(key)) {
@@ -90,10 +100,54 @@ function runJSify(functionsOnly) {
     });
   }
 
-  function processLibraryFunction(snippet, ident, finalName) {
-    // It is possible that when printing the function as a string on Windows, the js interpreter we are in returns the string with Windows
-    // line endings \r\n. This is undesirable, since line endings are managed in the form \n in the output for binary file writes, so
-    // make sure the endings are uniform.
+  function convertPointerParams(snippet, sig) {
+    // Automatically convert any incoming pointer arguments from BigInt
+    // to double (this limits the range to int53).
+    // And convert the return value if the function returns a pointer.
+    return modifyFunction(snippet, (name, args, body) => {
+      let argLines = args.split('\n');
+      argLines = argLines.map((line) => line.split('//')[0]);
+      const argNames = argLines.join(' ').split(',').map((name) => name.trim());
+      let newArgs = [];
+      let argConvertions = '';
+      for (let i = 1; i < sig.length; i++) {
+        const name = argNames[i - 1];
+        if (sig[i] == 'p') {
+          argConvertions += `  ${name} = Number(${name});\n`;
+          newArgs.push(`Number(${name})`);
+        } else {
+          newArgs.push(name);
+        }
+      }
+
+      if (sig[0] == 'p') {
+        // For functions that return a pointer we need to convert
+        // the return value too, which means we need to wrap the
+        // body in an inner function.
+        newArgs = newArgs.join(',');
+        return `\
+function ${name}(${args}) {
+  var ret = ((${args}) => { ${body} })(${newArgs});
+  return BigInt(ret);
+}`;
+      }
+
+      // Otherwise no inner function is needed and we covert the arguments
+      // before executing the function body.
+      return `\
+function ${name}(${args}) {
+${argConvertions}
+  ${body};
+}`;
+    });
+  }
+
+  function processLibraryFunction(snippet, ident, finalName, deps) {
+    // It is possible that when printing the function as a string on Windows,
+    // the js interpreter we are in returns the string with Windows line endings
+    // \r\n. This is undesirable, since line endings are managed in the form \n
+    // in the output for binary file writes, so make sure the endings are
+    // uniform.
     snippet = snippet.toString().replace(/\r\n/gm, '\n');
 
     // name the function; overwrite if it's already named
@@ -110,14 +164,42 @@ function ${name}(${args}) {
   return ret;
 }`);
     }
+
+    if (MEMORY64) {
+      const sig = LibraryManager.library[ident + '__sig'];
+      if (sig && sig.includes('p')) {
+        snippet = convertPointerParams(snippet, sig);
+      }
+    }
+
     return snippet;
   }
 
+  function addImplicitDeps(snippet, deps) {
+    // There are some common dependencies that we inject automatically by
+    // conservatively scanning the input functions for their usage.
+    // Specifically, these are dependencies that are automatically generated by
+    // the {{{ makeDynCall }}}, and {{{ runtimeKeepalivePush/Pop }}} macros.
+    const autoDeps = [
+      'getDynCaller',
+      'getWasmTableEntry',
+      'runtimeKeepalivePush',
+      'runtimeKeepalivePop',
+    ];
+    for (const dep of autoDeps) {
+      if (snippet.includes(dep + '(')) {
+        deps.push('$' + dep);
+      }
+    }
+  }
+
   // functionStub
   function functionStubHandler(item) {
-    // In LLVM, exceptions generate a set of functions of form __cxa_find_matching_catch_1(), __cxa_find_matching_catch_2(), etc.
-    // where the number specifies the number of arguments. In Emscripten, route all these to a single function '__cxa_find_matching_catch'
-    // that variadically processes all of these functions using JS 'arguments' object.
+    // In LLVM, exceptions generate a set of functions of form
+    // __cxa_find_matching_catch_1(), __cxa_find_matching_catch_2(), etc.  where
+    // the number specifies the number of arguments. In Emscripten, route all
+    // these to a single function '__cxa_find_matching_catch' that variadically
+    // processes all of these functions using JS 'arguments' object.
     if (item.identMangled.startsWith('___cxa_find_matching_catch_')) {
       if (DISABLE_EXCEPTION_THROWING) {
         error('DISABLE_EXCEPTION_THROWING was set (likely due to -fno-exceptions), which means no C++ exception throwing support code is linked in, but exception catching code appears. Either do not set DISABLE_EXCEPTION_THROWING (if you do want exception throwing) or compile all source files with -fno-except (so that no exceptions support code is required); also make sure DISABLE_EXCEPTION_CATCHING is set to the right value - if you want exceptions, it should be off, and vice versa.');
@@ -151,14 +233,26 @@ function ${name}(${args}) {
         return '';
       }
 
-      let noExport = false;
+      // This gets set to true in the case of dynamic linking for symbols that
+      // are undefined in the main module.  In this case we create a stub that
+      // will resolve the correct symbol at runtime, or assert if its missing.
+      let isStub = false;
 
       if (!LibraryManager.library.hasOwnProperty(ident)) {
         if (ONLY_CALC_JS_SYMBOLS) {
           return;
         }
-        if (!isDefined(ident)) {
-          let msg = 'undefined symbol: ' + ident;
+        const isWeakImport = WEAK_IMPORTS.has(ident);
+        if (!isDefined(ident) && !isWeakImport) {
+          if (PROXY_TO_PTHREAD && !MAIN_MODULE && ident == '__main_argc_argv') {
+            error('PROXY_TO_PTHREAD proxies main() for you, but no main exists');
+            return;
+          }
+          let undefinedSym = ident;
+          if (ident === '__main_argc_argv') {
+            undefinedSym = 'main/__main_argc_argv';
+          }
+          let msg = 'undefined symbol: ' + undefinedSym;
           if (dependent) msg += ` (referenced by ${dependent})`;
           if (ERROR_ON_UNDEFINED_SYMBOLS) {
             error(msg);
@@ -170,7 +264,7 @@ function ${name}(${args}) {
           } else if (VERBOSE || WARN_ON_UNDEFINED_SYMBOLS) {
             warn(msg);
           }
-          if (ident === 'main' && STANDALONE_WASM) {
+          if (ident === '__main_argc_argv' && STANDALONE_WASM) {
             warn('To build in STANDALONE_WASM mode without a main(), use emcc --no-entry');
           }
         }
@@ -189,13 +283,12 @@ function ${name}(${args}) {
           }
           const functionBody = assertion + `return ${target}.apply(null, arguments);`;
           LibraryManager.library[ident] = new Function(functionBody);
-          noExport = true;
+          isStub = true;
         }
       }
 
       const original = LibraryManager.library[ident];
       let snippet = original;
-      let redirectedIdent = null;
       const deps = LibraryManager.library[ident + '__deps'] || [];
       if (!Array.isArray(deps)) {
         error(`JS library directive ${ident}__deps=${deps.toString()} is of type ${typeof deps}, but it should be an array!`);
@@ -218,29 +311,21 @@ function ${name}(${args}) {
           if (target) {
             // Redirection for aliases. We include the parent, and at runtime make ourselves equal to it.
             // This avoid having duplicate functions with identical content.
-            redirectedIdent = snippet;
-            deps.push(snippet);
-            snippet = mangleCSymbolName(snippet);
-          }
-          // In asm, we need to know about library functions. If there is a target, though, then no
-          // need to consider this a library function - we will call directly to it anyhow
-          if (!redirectedIdent && (typeof target == 'function')) {
-            libraryFunctions.push(finalName);
+            const redirectedTarget = snippet;
+            deps.push(redirectedTarget);
+            snippet = mangleCSymbolName(redirectedTarget);
           }
         }
       } else if (typeof snippet == 'object') {
         snippet = stringifyWithFunctions(snippet);
+        addImplicitDeps(snippet, deps);
       } else if (typeof snippet == 'function') {
         isFunction = true;
-        snippet = processLibraryFunction(snippet, ident, finalName);
-        libraryFunctions.push(finalName);
+        snippet = processLibraryFunction(snippet, ident, finalName, deps);
+        addImplicitDeps(snippet, deps);
       }
 
-      // If a JS library item specifies xxx_import: true, then explicitly mark that symbol to be exported
-      // to wasm module.
-      if (LibraryManager.library[ident + '__import']) {
-        libraryFunctions.push(finalName);
-      }
+      librarySymbols.push(finalName);
 
       if (ONLY_CALC_JS_SYMBOLS) {
         return '';
@@ -262,9 +347,6 @@ function ${name}(${args}) {
         }
       }
 
-      if (redirectedIdent) {
-        deps = deps.concat(LibraryManager.library[redirectedIdent + '__deps'] || []);
-      }
       if (VERBOSE) {
         printErr(`adding ${finalName} and deps ${deps} : ` + (snippet + '').substr(0, 40));
       }
@@ -292,14 +374,19 @@ function ${name}(${args}) {
     return _emscripten_proxy_to_main_thread_js(${proxiedFunctionTable.length}, ${+sync}${args ? ', ' : ''}${args});
   ${body}
 }\n`);
-          } else if (WASM_WORKERS && ASSERTIONS) {
-            // In ASSERTIONS builds add runtime checks that proxied functions are not attempted to be called in Wasm Workers
-            // (since there is no automatic proxying architecture available)
-            contentText = modifyFunction(snippet, (name, args, body) => `
-function ${name}(${args}) {
-  assert(!ENVIRONMENT_IS_WASM_WORKER, "Attempted to call proxied function '${name}' in a Wasm Worker, but in Wasm Worker enabled builds, proxied function architecture is not available!");
-  ${body}
-}\n`);
+          } else if (WASM_WORKERS) {
+            if (ASSERTIONS) {
+              // In ASSERTIONS builds add runtime checks that proxied functions are not attempted to be called in Wasm Workers
+              // (since there is no automatic proxying architecture available)
+              contentText = modifyFunction(snippet, (name, args, body) => `
+  function ${name}(${args}) {
+    assert(!ENVIRONMENT_IS_WASM_WORKER, "Attempted to call proxied function '${name}' in a Wasm Worker, but in Wasm Worker enabled builds, proxied function architecture is not available!");
+    ${body}
+  }\n`);
+            } else {
+              // In non-ASSERTIONS builds directly emit the original function.
+              contentText = snippet;
+            }
           }
           proxiedFunctionTable.push(finalName);
         } else if ((USE_ASAN || USE_LSAN || UBSAN_RUNTIME) && LibraryManager.library[ident + '__noleakcheck']) {
@@ -333,12 +420,20 @@ function ${name}(${args}) {
       const sig = LibraryManager.library[ident + '__sig'];
       // asm module exports are done in emscripten.py, after the asm module is ready. Here
       // we also export library methods as necessary.
-      if ((EXPORT_ALL || EXPORTED_FUNCTIONS.has(finalName)) && !noExport) {
+      if ((EXPORT_ALL || EXPORTED_FUNCTIONS.has(finalName)) && !isStub) {
         contentText += `\nModule["${finalName}"] = ${finalName};`;
       }
-      if (MAIN_MODULE && sig) {
+      // Relocatable code needs signatures to create proper wrappers. Stack
+      // switching needs signatures so we can create a proper
+      // WebAssembly.Function with the signature for the Promise API.
+      // TODO: For asyncify we could only add the signatures we actually need,
+      //       of async imports/exports.
+      if (sig && (RELOCATABLE || ASYNCIFY == 2)) {
         contentText += `\n${finalName}.sig = '${sig}';`;
       }
+      if (isStub) {
+        contentText += `\n${finalName}.stub = true;`;
+      }
 
       let commentText = '';
       if (LibraryManager.library[ident + '__docs']) {
@@ -399,16 +494,11 @@ function ${name}(${args}) {
     print(pre);
 
     // Print out global variables and postsets TODO: batching
-    const legalizedI64sDefault = legalizedI64s;
-    legalizedI64s = false;
-
     runJSify(true);
 
     const generated = itemsDict.functionStub.concat(itemsDict.globalVariablePostSet);
     generated.forEach((item) => print(indentify(item.JS || '', 2)));
 
-    legalizedI64s = legalizedI64sDefault;
-
     if (USE_PTHREADS) {
       print('\n // proxiedFunctionTable specifies the list of functions that can be called either synchronously or asynchronously from other threads in postMessage()d or internally queued events. This way a pthread in a Worker can synchronously access e.g. the DOM on the main thread.');
       print('\nvar proxiedFunctionTable = [' + proxiedFunctionTable.join() + '];\n');
@@ -416,8 +506,6 @@ function ${name}(${args}) {
 
     if (!MINIMAL_RUNTIME) {
       print('var ASSERTIONS = ' + !!ASSERTIONS + ';\n');
-
-      print(preprocess(read('arrayUtils.js')));
     }
 
     if ((SUPPORT_BASE64_EMBEDDING || FORCE_FILESYSTEM) && !MINIMAL_RUNTIME) {
@@ -455,7 +543,8 @@ function ${name}(${args}) {
     print(processMacros(preprocess(shellParts[1], shellFile)));
 
     print('\n//FORWARDED_DATA:' + JSON.stringify({
-      libraryFunctions: libraryFunctions,
+      librarySymbols: librarySymbols,
+      warnings: warnings,
       ATINITS: ATINITS.join('\n'),
       ATMAINS: ATMAINS.join('\n'),
       ATEXITS: ATEXITS.join('\n'),
