@@ -5,6 +5,7 @@
  */
 
 var LibraryExceptions = {
+#if !WASM_EXCEPTIONS
   $uncaughtExceptionCount: '0',
   $exceptionLast: '0',
   $exceptionCaught: ' []',
@@ -72,7 +73,7 @@ var LibraryExceptions = {
     // Initialize native structure fields. Should be called once after allocated.
     this.init = function(type, destructor) {
 #if EXCEPTION_DEBUG
-      err('ExceptionInfo init: ' + [type, destructor]);
+      dbg('ExceptionInfo init: ' + [type, destructor]);
 #endif
       this.set_adjusted_ptr(0);
       this.set_type(type);
@@ -132,7 +133,7 @@ var LibraryExceptions = {
 
   $exception_addRef: function (info) {
 #if EXCEPTION_DEBUG
-    err('exception_addRef ' + ptrToString(info.excPtr));
+    dbg('exception_addRef ' + ptrToString(info.excPtr));
 #endif
     info.add_ref();
   },
@@ -144,7 +145,7 @@ var LibraryExceptions = {
   ],
   $exception_decRef: function(info) {
 #if EXCEPTION_DEBUG
-    err('exception_decRef ' + ptrToString(info.excPtr));
+    dbg('exception_decRef ' + ptrToString(info.excPtr));
 #endif
     // A rethrown exception can reach refcount 0; it must not be discarded
     // Its next handler will clear the rethrown flag and addRef it, prior to
@@ -153,24 +154,24 @@ var LibraryExceptions = {
       var destructor = info.get_destructor();
       if (destructor) {
         // In Wasm, destructors return 'this' as in ARM
-        {{{ makeDynCall('ii', 'destructor') }}}(info.excPtr);
+        {{{ makeDynCall('pp', 'destructor') }}}(info.excPtr);
       }
       ___cxa_free_exception(info.excPtr);
 #if EXCEPTION_DEBUG
-      err('decref freeing exception ' + [ptrToString(info.excPtr), exceptionLast, 'stack', exceptionCaught]);
+      dbg('decref freeing exception ' + [ptrToString(info.excPtr), exceptionLast, 'stack', exceptionCaught]);
 #endif
     }
   },
 
   // Exceptions
-  __cxa_allocate_exception__sig: 'vi',
+  __cxa_allocate_exception__sig: 'pp',
   __cxa_allocate_exception: function(size) {
     // Thrown object is prepended by exception metadata block
     return _malloc(size + {{{ C_STRUCTS.__cxa_exception.__size__ }}}) + {{{ C_STRUCTS.__cxa_exception.__size__ }}};
   },
 
   __cxa_free_exception__deps: ['$ExceptionInfo'],
-  __cxa_free_exception__sig: 'vi',
+  __cxa_free_exception__sig: 'vp',
   __cxa_free_exception: function(ptr) {
 #if ABORTING_MALLOC || ASSERTIONS
     try {
@@ -186,12 +187,14 @@ var LibraryExceptions = {
   },
 
   __cxa_increment_exception_refcount__deps: ['$exception_addRef', '$ExceptionInfo'],
+  __cxa_increment_exception_refcount__sig: 'vp',
   __cxa_increment_exception_refcount: function(ptr) {
     if (!ptr) return;
     exception_addRef(new ExceptionInfo(ptr));
   },
 
   __cxa_decrement_exception_refcount__deps: ['$exception_decRef', '$ExceptionInfo'],
+  __cxa_decrement_exception_refcount__sig: 'vp',
   __cxa_decrement_exception_refcount: function(ptr) {
     if (!ptr) return;
     exception_decRef(new ExceptionInfo(ptr));
@@ -199,11 +202,11 @@ var LibraryExceptions = {
 
   // Here, we throw an exception after recording a couple of values that we need to remember
   // We also remember that it was the last exception thrown as we need to know that later.
-  __cxa_throw__sig: 'viii',
+  __cxa_throw__sig: 'vppp',
   __cxa_throw__deps: ['$ExceptionInfo', '$exceptionLast', '$uncaughtExceptionCount'],
   __cxa_throw: function(ptr, type, destructor) {
 #if EXCEPTION_DEBUG
-    err('__cxa_throw: ' + [ptrToString(ptr), type, ptrToString(destructor)]);
+    dbg('__cxa_throw: ' + [ptrToString(ptr), type, ptrToString(destructor)]);
 #endif
     var info = new ExceptionInfo(ptr);
     // Initialize ExceptionInfo content after it was allocated in __cxa_allocate_exception.
@@ -232,18 +235,20 @@ var LibraryExceptions = {
       uncaughtExceptionCount++;
     }
 #if EXCEPTION_DEBUG
-    err('__cxa_rethrow, popped ' +
+    dbg('__cxa_rethrow, popped ' +
       [ptrToString(ptr), exceptionLast, 'stack', exceptionCaught]);
 #endif
     exceptionLast = ptr;
     {{{ makeThrow('ptr') }}}
   },
 
+  llvm_eh_typeid_for__sig: 'ip',
   llvm_eh_typeid_for: function(type) {
     return type;
   },
 
   __cxa_begin_catch__deps: ['$exceptionCaught', '$exception_addRef', '$uncaughtExceptionCount'],
+  __cxa_begin_catch__sig: 'pp',
   __cxa_begin_catch: function(ptr) {
     var info = new ExceptionInfo(ptr);
     if (!info.get_caught()) {
@@ -253,7 +258,7 @@ var LibraryExceptions = {
     info.set_rethrown(false);
     exceptionCaught.push(info);
 #if EXCEPTION_DEBUG
-    err('__cxa_begin_catch ' + [ptrToString(ptr), 'stack', exceptionCaught]);
+    dbg('__cxa_begin_catch ' + [ptrToString(ptr), 'stack', exceptionCaught]);
 #endif
     exception_addRef(info);
     return info.get_exception_ptr();
@@ -275,16 +280,17 @@ var LibraryExceptions = {
     var info = exceptionCaught.pop();
 
 #if EXCEPTION_DEBUG
-    err('__cxa_end_catch popped ' + [info, exceptionLast, 'stack', exceptionCaught]);
+    dbg('__cxa_end_catch popped ' + [info, exceptionLast, 'stack', exceptionCaught]);
 #endif
     exception_decRef(info);
     exceptionLast = 0; // XXX in decRef?
   },
 
   __cxa_get_exception_ptr__deps: ['$ExceptionInfo'],
+  __cxa_get_exception_ptr__sig: 'pp',
   __cxa_get_exception_ptr: function(ptr) {
 #if EXCEPTION_DEBUG
-    err('__cxa_get_exception_ptr ' + ptrToString(ptr));
+    dbg('__cxa_get_exception_ptr ' + ptrToString(ptr));
 #endif
     return new ExceptionInfo(ptr).get_exception_ptr();
   },
@@ -330,32 +336,34 @@ var LibraryExceptions = {
   // unwinding using 'if' blocks around each function, so the remaining
   // functionality boils down to picking a suitable 'catch' block.
   // We'll do that here, instead, to keep things simpler.
-  __cxa_find_matching_catch__deps: ['$exceptionLast', '$ExceptionInfo', '__resumeException', '__cxa_can_catch'],
+  __cxa_find_matching_catch__deps: ['$exceptionLast', '$ExceptionInfo', '__resumeException', '__cxa_can_catch', 'setTempRet0'],
+  //__cxa_find_matching_catch__sig: 'p',
   __cxa_find_matching_catch: function() {
     var thrown = exceptionLast;
     if (!thrown) {
       // just pass through the null ptr
-      {{{ makeStructuralReturn([0, 0]) }}};
+      setTempRet0(0);
+      return 0;
     }
     var info = new ExceptionInfo(thrown);
     info.set_adjusted_ptr(thrown);
     var thrownType = info.get_type();
     if (!thrownType) {
       // just pass through the thrown ptr
-      {{{ makeStructuralReturn(['thrown', 0]) }}};
+      setTempRet0(0);
+      return thrown;
     }
-    var typeArray = Array.prototype.slice.call(arguments);
 
     // can_catch receives a **, add indirection
 #if EXCEPTION_DEBUG
-    err("__cxa_find_matching_catch on " + ptrToString(thrown));
+    dbg("__cxa_find_matching_catch on " + ptrToString(thrown));
 #endif
     // The different catch blocks are denoted by different types.
     // Due to inheritance, those types may not precisely match the
     // type of the thrown object. Find one which matches, and
     // return the type of the catch block which should be called.
-    for (var i = 0; i < typeArray.length; i++) {
-      var caughtType = typeArray[i];
+    for (var i = 0; i < arguments.length; i++) {
+      var caughtType = arguments[i];
       if (caughtType === 0 || caughtType === thrownType) {
         // Catch all clause matched or exactly the same type is caught
         break;
@@ -363,31 +371,130 @@ var LibraryExceptions = {
       var adjusted_ptr_addr = info.ptr + {{{ C_STRUCTS.__cxa_exception.adjustedPtr }}};
       if ({{{ exportedAsmFunc('___cxa_can_catch') }}}(caughtType, thrownType, adjusted_ptr_addr)) {
 #if EXCEPTION_DEBUG
-        err("  __cxa_find_matching_catch found " + [ptrToString(info.get_adjusted_ptr()), caughtType]);
+        dbg("  __cxa_find_matching_catch found " + [ptrToString(info.get_adjusted_ptr()), caughtType]);
 #endif
-        {{{ makeStructuralReturn(['thrown', 'caughtType']) }}};
+        setTempRet0(caughtType);
+        return thrown;
       }
     }
-    {{{ makeStructuralReturn(['thrown', 'thrownType']) }}};
+    setTempRet0(thrownType);
+    return thrown;
   },
 
   __resumeException__deps: ['$exceptionLast'],
+  __resumeException__sig: 'vp',
   __resumeException: function(ptr) {
 #if EXCEPTION_DEBUG
-    err("__resumeException " + [ptrToString(ptr), exceptionLast]);
+    dbg("__resumeException " + [ptrToString(ptr), exceptionLast]);
 #endif
     if (!exceptionLast) { exceptionLast = ptr; }
     {{{ makeThrow('ptr') }}}
   },
 
-#if !DISABLE_EXCEPTION_CATCHING
-  $formatException__deps: ['emscripten_format_exception', 'free'],
-  $formatException: function(excPtr) {
-    var utf8_addr = _emscripten_format_exception(excPtr);
-    var result = UTF8ToString(utf8_addr);
-    _free(utf8_addr);
-    return result;
+#endif
+#if WASM_EXCEPTIONS || !DISABLE_EXCEPTION_CATCHING
+  $getExceptionMessageCommon__deps: ['__get_exception_message', 'free', '$withStackSave'],
+  $getExceptionMessageCommon: function(ptr) {
+    return withStackSave(function() {
+      var type_addr_addr = stackAlloc({{{ POINTER_SIZE }}});
+      var message_addr_addr = stackAlloc({{{ POINTER_SIZE }}});
+      ___get_exception_message({{{ to64('ptr') }}}, {{{ to64('type_addr_addr') }}}, {{{ to64('message_addr_addr') }}});
+      var type_addr = {{{ makeGetValue('type_addr_addr', 0, '*') }}};
+      var message_addr = {{{ makeGetValue('message_addr_addr', 0, '*') }}};
+      var type = UTF8ToString(type_addr);
+      _free(type_addr);
+      var message;
+      if (message_addr) {
+        message = UTF8ToString(message_addr);
+        _free(message_addr);
+      }
+      return [type, message];
+    });
   },
+#endif
+#if WASM_EXCEPTIONS
+  $getCppExceptionTag: function() {
+    // In static linking, tags are defined within the wasm module and are
+    // exported, whereas in dynamic linking, tags are defined in library.js in
+    // JS code and wasm modules import them.
+#if RELOCATABLE
+    return ___cpp_exception; // defined in library.js
+#else
+    return Module['asm']['__cpp_exception'];
+#endif
+  },
+
+#if ASSERTIONS
+  // Throw a WebAssembly.Exception object with the C++ tag with a stack trace
+  // embedded. WebAssembly.Exception is a JS object representing a Wasm
+  // exception, provided by Wasm JS API:
+  // https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/WebAssembly/Exception
+  // In release builds, this function is not needed and the native
+  // _Unwind_RaiseException in libunwind is used instead.
+  __throw_exception_with_stack_trace__deps: ['$getCppExceptionTag', '$getExceptionMessage'],
+  __throw_exception_with_stack_trace: function(ex) {
+    var e = new WebAssembly.Exception(getCppExceptionTag(), [ex], {traceStack: true});
+    e.message = getExceptionMessage(e);
+    // The generated stack trace will be in the form of:
+    //
+    // Error
+    //     at ___throw_exception_with_stack_trace(test.js:1139:13)
+    //     at __cxa_throw (wasm://wasm/009a7c9a:wasm-function[1551]:0x24367)
+    //     ...
+    //
+    // Remove this JS function name, which is in the second line, from the stack
+    // trace.
+    var arr = e.stack.split('\n');
+    arr.splice(1,1);
+    e.stack = arr.join('\n');
+    throw e;
+  },
+#endif
+
+  // Given an WebAssembly.Exception object, returns the actual user-thrown
+  // C++ object address in the Wasm memory.
+  $getCppExceptionThrownObjectFromWebAssemblyException__deps: ['$getCppExceptionTag', '__thrown_object_from_unwind_exception'],
+  $getCppExceptionThrownObjectFromWebAssemblyException: function(ex) {
+    // In Wasm EH, the value extracted from WebAssembly.Exception is a pointer
+    // to the unwind header. Convert it to the actual thrown value.
+    var unwind_header = ex.getArg(getCppExceptionTag(), 0);
+    return ___thrown_object_from_unwind_exception(unwind_header);
+  },
+
+  $incrementExceptionRefcount__deps: ['__cxa_increment_exception_refcount', '$getCppExceptionThrownObjectFromWebAssemblyException'],
+  $incrementExceptionRefcount: function(ex) {
+    var ptr = getCppExceptionThrownObjectFromWebAssemblyException(ex);
+    ___cxa_increment_exception_refcount(ptr);
+  },
+
+  $decrementExceptionRefcount__deps: ['__cxa_decrement_exception_refcount', '$getCppExceptionThrownObjectFromWebAssemblyException'],
+  $decrementExceptionRefcount: function(ex) {
+    var ptr = getCppExceptionThrownObjectFromWebAssemblyException(ex);
+    ___cxa_decrement_exception_refcount(ptr);
+  },
+
+  $getExceptionMessage__deps: ['$getCppExceptionThrownObjectFromWebAssemblyException', '$getExceptionMessageCommon'],
+  $getExceptionMessage: function(ex) {
+    var ptr = getCppExceptionThrownObjectFromWebAssemblyException(ex);
+    return getExceptionMessageCommon(ptr);
+  },
+
+#elif !DISABLE_EXCEPTION_CATCHING
+  $incrementExceptionRefcount__deps: ['__cxa_increment_exception_refcount'],
+  $incrementExceptionRefcount: function(ptr) {
+    ___cxa_increment_exception_refcount(ptr);
+  },
+
+  $decrementExceptionRefcount__deps: ['__cxa_decrement_exception_refcount'],
+  $decrementExceptionRefcount: function(ptr) {
+    ___cxa_decrement_exception_refcount(ptr);
+  },
+  $getExceptionMessage__deps: ['$getExceptionMessageCommon'],
+  $getExceptionMessage: function(ptr) {
+    return getExceptionMessageCommon(ptr);
+  },
+
+
 #endif
 };
 
@@ -396,7 +503,7 @@ var LibraryExceptions = {
 // that variadically processes all of these functions using JS 'arguments' object.
 addCxaCatch = function(n) {
   LibraryManager.library['__cxa_find_matching_catch_' + n] = LibraryExceptions['__cxa_find_matching_catch'];
-  LibraryManager.library['__cxa_find_matching_catch_' + n + '__sig'] = new Array(n + 2).join('i');
+  LibraryManager.library['__cxa_find_matching_catch_' + n + '__sig'] = new Array(n + 2).join('p');
   LibraryManager.library['__cxa_find_matching_catch_' + n + '__deps'] = LibraryExceptions['__cxa_find_matching_catch__deps'];
 };
 
