@@ -17,11 +17,11 @@ function genArgSequence(n) {
   return args;
 }
 
-// List of functions that were added from the library.
-global.libraryFunctions = [];
+// List of symbols that were added from the library.
+global.librarySymbols = [];
 
 global.LibraryManager = {
-  library: null,
+  library: {},
   structs: {},
   loaded: false,
   libraries: [],
@@ -31,38 +31,38 @@ global.LibraryManager = {
   },
 
   load: function() {
-    if (this.library) return;
+    assert(!this.loaded);
+    this.loaded = true;
 
     // Core system libraries (always linked against)
     let libraries = [
       'library.js',
+      'library_int53.js',
+      'library_ccall.js',
+      'library_addfunction.js',
       'library_formatString.js',
+      'library_getvalue.js',
       'library_math.js',
       'library_path.js',
+      'library_strings.js',
       'library_syscall.js',
       'library_html5.js',
       'library_stack_trace.js',
       'library_wasi.js',
-      'library_int53.js',
       'library_dylink.js',
+      'library_makeDynCall.js',
       'library_eventloop.js',
     ];
 
-    if (LINK_AS_CXX && !EXCEPTION_HANDLING) {
-      if (DISABLE_EXCEPTION_THROWING) {
+    if (LINK_AS_CXX) {
+      if (DISABLE_EXCEPTION_THROWING && !WASM_EXCEPTIONS) {
         libraries.push('library_exceptions_stub.js');
       } else {
         libraries.push('library_exceptions.js');
       }
     }
 
-    if (MINIMAL_RUNTIME) {
-      // Classic runtime brings in string-related functions in the runtime preamble, by including
-      // runtime_strings_extra.js that contain the same contents as library_strings.js does. In
-      // MINIMAL_RUNTIME those string functions are available as JS library functions instead from
-      // library_strings.js, to avoid unconditionally bringing in extra code to the build.
-      libraries.push('library_strings.js');
-    } else {
+    if (!MINIMAL_RUNTIME) {
       libraries.push('library_browser.js');
       libraries.push('library_wget.js');
     }
@@ -71,6 +71,10 @@ global.LibraryManager = {
       libraries.push('library_memoryprofiler.js');
     }
 
+    if (AUTODEBUG) {
+      libraries.push('library_autodebug.js');
+    }
+
     if (FILESYSTEM) {
       // Core filesystem libraries (always linked against, unless -sFILESYSTEM=0 is specified)
       libraries = libraries.concat([
@@ -87,12 +91,16 @@ global.LibraryManager = {
           libraries.push('library_nodefs.js');
         }
         libraries.push('library_noderawfs.js');
+        // NODERAWFS overwrites library_path.js
+        libraries.push('library_nodepath.js');
       }
     } else if (WASMFS) {
       libraries.push('library_wasmfs.js');
       libraries.push('library_wasmfs_js_file.js');
+      libraries.push('library_wasmfs_jsimpl.js');
       libraries.push('library_wasmfs_fetch.js');
       libraries.push('library_wasmfs_node.js');
+      libraries.push('library_wasmfs_opfs.js');
     }
 
     // Additional JS libraries (without AUTO_JS_LIBRARIES, link to these explicitly via -lxxx.js)
@@ -147,10 +155,15 @@ global.LibraryManager = {
       libraries.push('library_html5_webgpu.js');
     }
 
+    if (!STRICT) {
+      libraries.push('library_legacy.js');
+    }
+
     if (BOOTSTRAPPING_STRUCT_INFO) {
       libraries = [
         'library_bootstrap.js',
         'library_formatString.js',
+        'library_strings.js',
         'library_int53.js',
       ];
     }
@@ -228,79 +241,15 @@ global.LibraryManager = {
       }
     }
 
-    // apply synonyms. these are typically not speed-sensitive, and doing it
-    // this way makes it possible to not include hacks in the compiler
-    // (and makes it simpler to switch between SDL versions, fastcomp and non-fastcomp, etc.).
-    const lib = this.library;
-    libloop: for (const x in lib) {
-      if (!Object.prototype.hasOwnProperty.call(lib, x)) {
-        continue;
-      }
-      if (isJsLibraryConfigIdentifier(x)) {
-        const index = x.lastIndexOf('__');
-        const basename = x.slice(0, index);
-        if (!(basename in lib)) {
-          error(`Missing library element '${basename}' for library config '${x}'`);
-        }
-        continue;
-      }
-      if (typeof lib[x] == 'string') {
-        let target = x;
-        while (typeof lib[target] == 'string') {
-          // ignore code and variable assignments, aliases are just simple names
-          if (lib[target].search(/[=({; ]/) >= 0) continue libloop;
-          target = lib[target];
-        }
-        if (!isNaN(target)) continue; // This is a number, and so cannot be an alias target.
-        if (typeof lib[target] == 'undefined' || typeof lib[target] == 'function') {
-          // When functions are aliased, a signature for the function must be
-          // provided so that an efficient form of forwarding can be
-          // implemented.
-          function testStringType(sig) {
-            if (typeof lib[sig] != 'undefined' && typeof typeof lib[sig] != 'string') {
-              error(`${sig} should be a string! (was ${typeof lib[sig]})`);
-            }
-          }
-          const aliasSig = x + '__sig';
-          const targetSig = target + '__sig';
-          testStringType(aliasSig);
-          testStringType(targetSig);
-          if (typeof lib[aliasSig] == 'string' && typeof lib[targetSig] == 'string' && lib[aliasSig] != lib[targetSig]) {
-            error(`${aliasSig} (${lib[aliasSig]}) differs from ${targetSig} (${lib[targetSig]})`);
-          }
-
-          const sig = lib[aliasSig] || lib[targetSig];
-          if (typeof sig != 'string') {
-            error(`Function ${x} aliases to target function ${target}, but neither the alias or the target provide a signature. Please add a ${targetSig}: 'vifj...' annotation or a ${aliasSig}: 'vifj...' annotation to describe the type of function forwarding that is needed!`);
-          }
-
-          // If only one of the target or the alias specifies a sig then copy
-          // this signature to the other.
-          if (!lib[aliasSig]) {
-            lib[aliasSig] = lib[targetSig];
-          } else if (!lib[targetSig]) {
-            lib[targetSig] = lib[aliasSig];
-          }
-
-          if (typeof lib[target] != 'function') {
-            error(`no alias found for ${x}`);
-          }
-
-          const argCount = sig.length - 1;
-          if (argCount !== lib[target].length) {
-            error(`incorrect number of arguments in signature of ${x} (declared: ${argCount}, expected: ${lib[target].length})`);
-          }
-          const ret = sig == 'v' ? '' : 'return ';
-          const args = genArgSequence(argCount).join(',');
-          lib[x] = new Function(args, `${ret}_${target}(${args});`);
-
-          if (!lib[x + '__deps']) lib[x + '__deps'] = [];
-          lib[x + '__deps'].push(target);
+    for (const ident in this.library) {
+      if (isJsLibraryConfigIdentifier(ident)) {
+        const index = ident.lastIndexOf('__');
+        const basename = ident.slice(0, index);
+        if (!(basename in this.library)) {
+          error(`Missing library element '${basename}' for library config '${ident}'`);
         }
       }
     }
-
-    this.loaded = true;
   },
 };
 
@@ -324,32 +273,56 @@ function isFSPrefixed(name) {
   return name.length > 3 && name[0] === 'F' && name[1] === 'S' && name[2] === '_';
 }
 
-// forcing the filesystem exports a few things by default
-function isExportedByForceFilesystem(name) {
-  if (!WASMFS) {
-    // The old FS has some functionality that WasmFS lacks.
-    if (name === 'FS_createLazyFile' ||
-        name === 'FS_createDevice') {
-      return true;
+function isInternalSymbol(ident) {
+  return ident + '__internal' in LibraryManager.library;
+}
+
+// When running with ASSERTIONS enabled we create stubs for each library
+// function that that was not included in the build.  This gives useful errors
+// when library dependencies are missing from `__deps` or depended on without
+// being added to DEFAULT_LIBRARY_FUNCS_TO_INCLUDE
+// TODO(sbc): These errors could potentially be generated at build time via
+// some kind of acorn pass that searched for uses of these missing symbols.
+function addMissingLibraryStubs() {
+  if (!ASSERTIONS) return '';
+  let rtn = '';
+  const librarySymbolSet = new Set(librarySymbols);
+  const missingSyms = [];
+  for (const ident in LibraryManager.library) {
+    if (typeof LibraryManager.library[ident] === 'function' || typeof LibraryManager.library[ident] === 'number') {
+      if (ident[0] === '$' && !isJsLibraryConfigIdentifier(ident) && !isInternalSymbol(ident)) {
+        const name = ident.substr(1);
+        if (!librarySymbolSet.has(name)) {
+          missingSyms.push(name);
+        }
+      }
     }
   }
-  return name === 'FS_createPath' ||
-         name === 'FS_createDataFile' ||
-         name === 'FS_createPreloadedFile' ||
-         name === 'FS_unlink' ||
-         name === 'addRunDependency' ||
-         name === 'removeRunDependency';
+  if (missingSyms.length) {
+    rtn += 'var missingLibrarySymbols = [\n';
+    for (const sym of missingSyms) {
+      rtn += `  '${sym}',\n`;
+    }
+    rtn += '];\n';
+    rtn += 'missingLibrarySymbols.forEach(missingLibrarySymbol)\n';
+  }
+  return rtn;
 }
 
 // export parts of the JS runtime that the user asked for
 function exportRuntime() {
   const EXPORTED_RUNTIME_METHODS_SET = new Set(EXPORTED_RUNTIME_METHODS);
 
+  const legacyRuntimeElements = new Map([
+    ['print', 'out'],
+    ['printErr', 'err'],
+  ]);
+
   // optionally export something.
   // in ASSERTIONS mode we show a useful error if it is used without
   // being exported. how we show the message depends on whether it's
   // a function (almost all of them) or a number.
-  function maybeExport(name, isNumber) {
+  function maybeExport(name) {
     // if requested to be exported, export it
     if (EXPORTED_RUNTIME_METHODS_SET.has(name)) {
       let exported = name;
@@ -357,55 +330,26 @@ function exportRuntime() {
       if (isFSPrefixed(exported)) {
         // this is a filesystem value, FS.x exported as FS_x
         exported = 'FS.' + exported.substr(3);
-      } else if (exported === 'print') {
-        exported = 'out';
-      } else if (exported === 'printErr') {
-        exported = 'err';
+      } else if (legacyRuntimeElements.has(exported)) {
+        exported = legacyRuntimeElements.get(exported);
       }
       return `Module["${name}"] = ${exported};`;
     }
-    // do not export it. but if ASSERTIONS, emit a
-    // stub with an error, so the user gets a message
-    // if it is used, that they should export it
-    if (ASSERTIONS) {
-      // check if it already exists, to support EXPORT_ALL and other cases
-      const fssymbol = isExportedByForceFilesystem(name);
-      if (isNumber) {
-        return `unexportedRuntimeSymbol('${name}', ${fssymbol});`;
-      } else {
-        return `unexportedRuntimeFunction('${name}', ${fssymbol});`;
-      }
-    }
-    return '';
-  }
-
-  function maybeExportNumber(name) {
-    return maybeExport(name, true);
   }
 
   // All possible runtime elements that can be exported
   let runtimeElements = [
-    'intArrayFromString',
-    'intArrayToString',
-    'ccall',
-    'cwrap',
-    'setValue',
-    'getValue',
-    'allocate',
+    'run',
     'UTF8ArrayToString',
     'UTF8ToString',
     'stringToUTF8Array',
     'stringToUTF8',
     'lengthBytesUTF8',
-    'stackTrace',
     'addOnPreRun',
     'addOnInit',
     'addOnPreMain',
     'addOnExit',
     'addOnPostRun',
-    'writeStringToMemory',
-    'writeArrayToMemory',
-    'writeAsciiToMemory',
     'addRunDependency',
     'removeRunDependency',
     'FS_createFolder',
@@ -420,20 +364,32 @@ function exportRuntime() {
     'getFunctionTables',
     'alignFunctionTables',
     'registerFunctions',
-    'addFunction',
-    'removeFunction',
     'prettyPrint',
-    'dynCall',
     'getCompilerSetting',
-    'print',
-    'printErr',
-    'getTempRet0',
-    'setTempRet0',
+    'out',
+    'err',
     'callMain',
     'abort',
     'keepRuntimeAlive',
+    'wasmMemory',
   ];
 
+  // These are actually native wasm functions these days but we allow exporting
+  // them via EXPORTED_RUNTIME_METHODS for backwards compat.
+  runtimeElements = runtimeElements.concat(WASM_SYSTEM_EXPORTS);
+
+  if (USE_PTHREADS && ALLOW_MEMORY_GROWTH) {
+    runtimeElements = runtimeElements.concat([
+      'GROWABLE_HEAP_I8',
+      'GROWABLE_HEAP_U8',
+      'GROWABLE_HEAP_I16',
+      'GROWABLE_HEAP_U16',
+      'GROWABLE_HEAP_I32',
+      'GROWABLE_HEAP_U32',
+      'GROWABLE_HEAP_F32',
+      'GROWABLE_HEAP_F64',
+    ]);
+  }
   if (USE_OFFSET_CONVERTER) {
     runtimeElements.push('WasmOffsetConverter');
   }
@@ -442,58 +398,16 @@ function exportRuntime() {
     runtimeElements.push('WasmSourceMap');
   }
 
-  // Add JS library elements such as FS, GL, ENV, etc. These are prefixed with
-  // '$ which indicates they are JS methods.
-  for (const ident in LibraryManager.library) {
-    if (ident[0] === '$' && !isJsLibraryConfigIdentifier(ident) && !LibraryManager.library[ident + '__internal']) {
-      runtimeElements.push(ident.substr(1));
-    }
-  }
-
-  if (!MINIMAL_RUNTIME) {
-    // MINIMAL_RUNTIME has moved these functions to library_strings.js
-    runtimeElements = runtimeElements.concat([
-      'warnOnce',
-      'stackSave',
-      'stackRestore',
-      'stackAlloc',
-      'AsciiToString',
-      'stringToAscii',
-      'UTF16ToString',
-      'stringToUTF16',
-      'lengthBytesUTF16',
-      'UTF32ToString',
-      'stringToUTF32',
-      'lengthBytesUTF32',
-      'allocateUTF8',
-      'allocateUTF8OnStack',
-    ]);
-  }
-
   if (STACK_OVERFLOW_CHECK) {
     runtimeElements.push('writeStackCookie');
     runtimeElements.push('checkStackCookie');
   }
 
-  if (USE_PTHREADS) {
-    // In pthreads mode, the following functions always need to be exported to
-    // Module for closure compiler, and also for MODULARIZE (so worker.js can
-    // access them).
-    const threadExports = ['PThread', 'wasmMemory'];
-    if (!MINIMAL_RUNTIME) {
-      threadExports.push('ExitStatus');
-    }
-
-    threadExports.forEach((x) => {
-      EXPORTED_RUNTIME_METHODS_SET.add(x);
-      runtimeElements.push(x);
-    });
-  }
-
   if (SUPPORT_BASE64_EMBEDDING) {
     runtimeElements.push('intArrayFromBase64');
     runtimeElements.push('tryParseAsDataURI');
   }
+
   // dynCall_* methods are not hardcoded here, as they
   // depend on the file being compiled. check for them
   // and add them.
@@ -503,20 +417,56 @@ function exportRuntime() {
       runtimeElements.push(name);
     }
   }
-  const runtimeNumbers = [
-    'ALLOC_NORMAL',
-    'ALLOC_STACK',
-  ];
-  if (ASSERTIONS) {
-    // check all exported things exist, warn about typos
-    for (const name of EXPORTED_RUNTIME_METHODS_SET) {
-      if (!runtimeElements.includes(name) && !runtimeNumbers.includes(name)) {
-        printErr(`warning: invalid item (maybe a typo?) in EXPORTED_RUNTIME_METHODS: ${name}`);
-      }
+
+  // Only export legacy runtime elements when explicitly
+  // requested.
+  for (const name of EXPORTED_RUNTIME_METHODS_SET) {
+    if (legacyRuntimeElements.has(name)) {
+      const newName = legacyRuntimeElements.get(name);
+      warn(`deprecated item in EXPORTED_RUNTIME_METHODS: ${name} use ${newName} instead.`);
+      runtimeElements.push(name);
     }
   }
+
+  // Add JS library elements such as FS, GL, ENV, etc. These are prefixed with
+  // '$ which indicates they are JS methods.
+  const runtimeElementsSet = new Set(runtimeElements);
+  for (const ident in LibraryManager.library) {
+    if (ident[0] === '$' && !isJsLibraryConfigIdentifier(ident) && !isInternalSymbol(ident)) {
+      const jsname = ident.substr(1);
+      assert(!runtimeElementsSet.has(jsname), 'runtimeElements contains library symbol: ' + ident);
+      runtimeElements.push(jsname);
+    }
+  }
+
+  let unexportedStubs = '';
+  if (ASSERTIONS) {
+    // check all exported things exist, warn about typos
+    const runtimeElementsSet = new Set(runtimeElements);
+    for (const name of EXPORTED_RUNTIME_METHODS_SET) {
+      if (!runtimeElementsSet.has(name)) {
+        warn(`invalid item in EXPORTED_RUNTIME_METHODS: ${name}`);
+      }
+    }
+
+    const unexported = [];
+    for (const name of runtimeElements) {
+      if (!EXPORTED_RUNTIME_METHODS_SET.has(name)) {
+        unexported.push(name);
+      }
+    }
+
+    if (unexported.length) {
+      unexportedStubs += 'var unexportedRuntimeSymbols = [\n';
+      for (const sym of unexported) {
+        unexportedStubs += `  '${sym}',\n`;
+      }
+      unexportedStubs += '];\n';
+      unexportedStubs += 'unexportedRuntimeSymbols.forEach(unexportedRuntimeSymbol);\n';
+    }
+  }
+
   let exports = runtimeElements.map((name) => maybeExport(name));
-  exports = exports.concat(runtimeNumbers.map((name) => maybeExportNumber(name)));
-  exports = exports.filter((name) => name != '');
-  return exports.join('\n');
+  exports = exports.filter((name) => name);
+  return exports.join('\n') + '\n' + unexportedStubs + addMissingLibraryStubs();
 }
