@@ -25,6 +25,7 @@
 #include "em_types.h"
 #include "em_js.h"
 #include "wget.h"
+#include "version.h"
 
 #ifdef __cplusplus
 extern "C" {
@@ -70,8 +71,8 @@ void emscripten_set_main_loop_expected_blockers(int num);
 
 void emscripten_async_call(em_arg_callback_func func, void *arg, int millis);
 
-void emscripten_exit_with_live_runtime(void) __attribute__((noreturn));
-void emscripten_force_exit(int status) __attribute__((noreturn));
+void emscripten_exit_with_live_runtime(void) __attribute__((__noreturn__));
+void emscripten_force_exit(int status) __attribute__((__noreturn__));
 
 double emscripten_get_device_pixel_ratio(void);
 
@@ -135,7 +136,7 @@ int emscripten_get_worker_queue_size(worker_handle worker);
 
 // misc.
 
-int emscripten_get_compiler_setting(const char *name);
+long emscripten_get_compiler_setting(const char *name);
 int emscripten_has_asyncify(void);
 
 void emscripten_debugger(void);
