@@ -27,11 +27,11 @@ def get(ports, settings, shared):
     ]
     ports.build_port(source_path, final, 'bzip2', srcs=srcs)
 
-  return [shared.Cache.get_lib('libbz2.a', create, what='port')]
+  return [shared.cache.get_lib('libbz2.a', create, what='port')]
 
 
 def clear(ports, settings, shared):
-  shared.Cache.erase_lib('libbz2.a')
+  shared.cache.erase_lib('libbz2.a')
 
 
 def process_args(ports):
