@@ -6,12 +6,14 @@
 import difflib
 import os
 import re
+from typing import Set, Dict, Any
 
 from .utils import path_from_root, exit_with_error
 from . import diagnostics
 
 # Subset of settings that take a memory size (i.e. 1Gb, 64kb etc)
 MEM_SIZE_SETTINGS = {
+    'STACK_SIZE',
     'TOTAL_STACK',
     'INITIAL_MEMORY',
     'MEMORY_GROWTH_LINEAR_STEP',
@@ -47,6 +49,7 @@ PORTS_SETTINGS = {
     'USE_FREETYPE',
     'SDL2_MIXER_FORMATS',
     'SDL2_IMAGE_FORMATS',
+    'USE_SQLITE3',
 }
 
 # Subset of settings that apply at compile time.
@@ -70,26 +73,26 @@ COMPILE_TIME_SETTINGS = {
 
     # Internal settings used during compilation
     'EXCEPTION_CATCHING_ALLOWED',
-    'EXCEPTION_HANDLING',
+    'WASM_EXCEPTIONS',
     'LTO',
     'OPT_LEVEL',
     'DEBUG_LEVEL',
 
     # This is legacy setting that we happen to handle very early on
     'RUNTIME_LINKED_LIBS',
-    # TODO: should not be here
-    'AUTO_ARCHIVE_INDEXES',
-    'DEFAULT_LIBRARY_FUNCS_TO_INCLUDE',
 }.union(PORTS_SETTINGS)
 
 
+user_settings: Dict[str, str] = {}
+
+
 class SettingsManager:
-  attrs = {}
-  types = {}
-  allowed_settings = set()
-  legacy_settings = {}
-  alt_names = {}
-  internal_settings = set()
+  attrs: Dict[str, Any] = {}
+  types: Dict[str, Any] = {}
+  allowed_settings: Set[str] = set()
+  legacy_settings: Dict[str, tuple] = {}
+  alt_names: Dict[str, str] = {}
+  internal_settings: Set[str] = set()
 
   def __init__(self):
     self.attrs.clear()
