@@ -34,11 +34,11 @@ typedef signed long ssize_t;
 
 #define TRAP(x) (wasm_rt_trap(WASM_RT_TRAP_##x), 0)
 
-#define MEMACCESS(addr) ((void*)&Z_memory->data[addr])
+#define MEMACCESS(addr) ((void*)&WASM_RT_ADD_PREFIX(Z_memory)->data[addr])
 
 #undef MEMCHECK
 #define MEMCHECK(a, t)  \
-  if (UNLIKELY((a) + sizeof(t) > Z_memory->size)) TRAP(OOB)
+  if (UNLIKELY((a) + sizeof(t) > WASM_RT_ADD_PREFIX(Z_memory)->size)) TRAP(OOB)
 
 #undef DEFINE_LOAD
 #define DEFINE_LOAD(name, t1, t2, t3)              \
@@ -94,7 +94,7 @@ static ret _##name params { \
   VERBOSE_LOG("[import: " #name "]\n"); \
   body \
 } \
-ret (*name) params = _##name;
+ret (*WASM_RT_ADD_PREFIX(name)) params = _##name;
 
 #define STUB_IMPORT_IMPL(ret, name, params, returncode) IMPORT_IMPL(ret, name, params, { return returncode; });
 
@@ -102,7 +102,7 @@ ret (*name) params = _##name;
 
 static void abort_with_message(const char* message) {
   fprintf(stderr, "%s\n", message);
-  abort();
+  TRAP(UNREACHABLE);
 }
 
 // Maintain a stack of setjmps, each jump taking us back to the last invoke.
@@ -113,40 +113,15 @@ static jmp_buf setjmp_stack[MAX_SETJMP_STACK];
 
 static u32 next_setjmp = 0;
 
-// Declare an export that may be needed and may not be. For example if longjmp
-// is included then we need setThrew, but we must declare setThrew so that
-// the C compiler can build us without error if longjmp is not actually used.
-
-#define DECLARE_WEAK_EXPORT(ret, name, args) \
-__attribute__((weak)) \
-ret (*WASM_RT_ADD_PREFIX(name)) args = NULL;
-
-DECLARE_WEAK_EXPORT(void, Z_setThrewZ_vii, (u32, u32));
-
-IMPORT_IMPL(void, Z_envZ_emscripten_longjmpZ_vii, (u32 buf, u32 value), {
+IMPORT_IMPL(void, Z_envZ__emscripten_throw_longjmpZ_vv, (), {
   if (next_setjmp == 0) {
     abort_with_message("longjmp without setjmp");
   }
-  Z_setThrewZ_vii(buf, value ? value : 1);
   longjmp(setjmp_stack[next_setjmp - 1], 1);
 });
 
-IMPORT_IMPL(void, Z_envZ_emscripten_longjmp_jmpbufZ_vii, (u32 buf, u32 value), {
-  return Z_envZ_emscripten_longjmpZ_vii(buf, value);
-});
-
 IMPORT_IMPL(void, Z_envZ_emscripten_notify_memory_growthZ_vi, (u32 size), {});
 
-static u32 tempRet0 = 0;
-
-IMPORT_IMPL(u32, Z_envZ_getTempRet0Z_iv, (), {
-  return tempRet0;
-});
-
-IMPORT_IMPL(void, Z_envZ_setTempRet0Z_vi, (u32 x), {
-  tempRet0 = x;
-});
-
 // Shared OS support in both sandboxed and unsandboxed mode
 
 #define WASI_DEFAULT_ERROR 63 /* __WASI_ERRNO_PERM */
